<?php
/**
 * Record Controller
 *
 * PHP version 5
 *
 * Copyright (C) Villanova University 2010.
 * Copyright (C) Leipzig University Library 2016.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2,
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * @category VuFind
 * @package  Controller
 * @author   Demian Katz <demian.katz@villanova.edu>
 * @author   André Lahmann <lahmann@ub.uni-leipzig.de>
 * @license  http://opensource.org/licenses/gpl-2.0.php GNU General Public License
 * @link     https://vufind.org Main Site
 */
namespace finc\Controller;

use finc\Rewrite\EblRewrite;
use VuFind\Exception\RecordMissing;
use Zend\Log\LoggerAwareInterface as LoggerAwareInterface;
use Zend\View\View;

/**
 * Record Controller
 *
 * @category VuFind
 * @package  Controller
 * @author   Demian Katz <demian.katz@villanova.edu>
 * @author   André Lahmann <lahmann@ub.uni-leipzig.de>
 * @license  http://opensource.org/licenses/gpl-2.0.php GNU General Public License
 * @link     https://vufind.org Main Site
 */
class RecordController extends \VuFind\Controller\RecordController implements
    LoggerAwareInterface
{
    use \VuFind\Log\LoggerAwareTrait;
    use CustomTraits\EblTrait;
    use CustomTraits\EmailHoldTrait;
    use CustomTraits\EmailProfileTrait;
    use CustomTraits\PdaTrait;
    use CustomTraits\ReportErrorsTrait;
    use HoldsTrait;

    /**
     * Returns rewrite object
     *
     * @return object
     */
    protected function getRewrite()
    {
        return $this->serviceLocator->get(EblRewrite::class);
    }

    /**
     * ProcessSave -- store the results of the Save action.
     *
     * @return mixed
     */
    protected function processSave()
    {
        // Retrieve user object and force login if necessary:
        if (!($user = $this->getUser())) {
            return $this->forceLogin();
        }

        // Perform the save operation:
        $driver = $this->loadRecord();
        $post = $this->getRequest()->getPost()->toArray();
        $tagParser = $this->serviceLocator->get('VuFind\Tags');
        $post['mytags']
            = $tagParser->parse($post['mytags'] ?? '');
        $favorites = $this->serviceLocator
            ->get('VuFind\Favorites\FavoritesService');
        $results = $favorites->save($post, $user, $driver);

        // Display a success status message:
        $listUrl = $this->url()->fromRoute('userList', ['id' => $results['listId']]);
        $message = [
            'html' => true,
            'msg' => $this->translate('bulk_save_success') . '. '
                . '<a href="' . $listUrl . '" class="gotolist">'
                . $this->translate('go_to_list') . '</a>.'
        ];
        $this->flashMessenger()->addMessage($message, 'success');

        // finc specific: dont redirect in layout context for performance reasons
        if ($this->params()->fromPost('layout', $this->params()->fromQuery('layout')) === 'lightbox') {
            $view = $this->createViewModel($this->params()->fromPost());
            $view->setTemplate('default/flash-message');
            return $view;
        }
        // finc specific - END

        // redirect to followup url saved in saveAction
        if ($url = $this->getFollowupUrl()) {
            $this->clearFollowupUrl();
            return $this->redirect()->toUrl($url);
        }

        // No followup info found?  Send back to record view:
        return $this->redirectToRecord();
    }

    /**
     * Home (default) action -- forward to requested (or default) tab.
     *
     * @return mixed
     */
    public function homeAction()
    {
        try {
            return parent::homeAction();
        } catch (\VuFind\Exception\RecordMissing $ex) {
            if (
                $this->params()->fromQuery('catchMissing', null)
                &&
                $lookfor = $this->params()->fromQuery('catchMissing_lookfor', '')
            ) {
                $type = $this->params()->fromQuery('catchMissing_searchType', 'Allfields');
                $link = $this->redirect()->toRoute(
                    'search-results',
                    [],
                    ['query' => compact('lookfor', 'type')]
                );
                $this->flashMessenger()->addErrorMessage('missing_record_redirect');
                return $link;
            } else {
                $id = $this->params()->fromRoute('id');
                throw new RecordMissing($id);
            }
        }

        return parent::homeAction();
    }
}
