<?php
/**
 * Finc Theca ILS driver
 *
 * PHP version 7
 *
 * Copyright (C) Leipzig University Library 2018.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2,
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * @category Finc
 * @package  FincILSDriver
 * @author   Dorian Merz <merz@ub.uni-leipzig.de>
 * @license  http://opensource.org/licenses/gpl-2.0.php GNU General Public License
 * @link     https://vufind.org Main Site
 */
namespace finc\ILS\Driver;

/**
 * Finc Theca ILS driver
 *
 * @category Finc
 * @package  FincILSDriver
 * @author   Dorian Merz <merz@ub.uni-leipzig.de>
 * @license  http://opensource.org/licenses/gpl-2.0.php GNU General Public License
 * @link     https://vufind.org Main Site
 * @deprecated Remove when Bibliotheca support ends
 */
class FincTheca extends FincILS
{
    private $identifier_type = 'ppn';
    private $daiaIdPrefixBase;

    /**
     * Init
     *
     * @return void
     * @access public
     */
    public function init()
    {
        parent::init();
        if (isset($this->config['DAIA']['daiaIdPrefixBase'])) {
            $this->daiaIdPrefixBase = $this->config['DAIA']['daiaIdPrefixBase'];
        }
    }

    /**
     * Get status
     *
     * @param string $id
     *
     * @return array|mixed
     * @access public
     */
    public function getStatus($id)
    {
        $result = [];

        try {
            $result = parent::getStatus($id);
        } catch (\Exception $e) {
        }

        if (empty($result)) {
            $result = $this->getStatusViaMediennummer($id);
        }
        return $result;
    }

    /**
     * @param string $record_id Finc-ID of record, used to retrieve Mediennummer
     *
     * @return array|mixed
     * @access protected
     */
    protected function getStatusViaMediennummer($record_id)
    {
        $result = [];
        if ($ilsRecordId = $this->getRecord($record_id)->tryMethod('getMediennummer')) {
            $this->identifier_type = 'mediennr';
            $result = parent::doGetStatus($ilsRecordId);
            foreach ($result as &$item) {
                //fix-up IDs
                $item['id'] = $record_id;
            }
        }
        return $result;
    }

    /**
     * Get statuses
     *
     * @param array $ids
     *
     * @return array
     * @access public
     */
    public function getStatuses($ids)
    {
        $results = parent::getStatuses($ids);
        if (count($results) < count($ids)) {
            //some records had no availability info
            $missing = array_flip($ids);
            foreach ($results as $items) {
                $item = current($items);
                if (isset($missing[$item['id']])) {
                    unset($missing[$item['id']]);
                }
            }
            foreach (array_keys($missing) as $missing_id) {
                $results[] = $this->getStatusViaMediennummer($missing_id);
            }
        }
        return $results;
    }

    /**
     * Generate uri
     *
     * @param string $id
     *
     * @return string
     * @access protected
     */
    protected function generateURI($id)
    {
        if (isset($this->daiaIdPrefixBase)) {
            return
                $this->daiaIdPrefixBase
                .':'.$this->identifier_type
                .':'.$id;
        }
        return parent::generateMultiURIs($id);
    }

    /**
     * Get item barcode
     *
     * @param array $item
     *
     * @return null
     * @access protected
     */
    protected function getItemBarcode($item)
    {
        $matches = [];
        if (preg_match(
            '/^' . $this->daiaIdPrefixBase . ':' . $this->identifier_type . ':\w+:(\w+)$/',
            $item['id'],
            $matches
        )) {
            return $matches[1];
        }
        return null;
    }
}
